#include "m_pd.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include<stdio.h>
#include <time.h>
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

/* ------------------------ fir~ ----------------------------- */
/* fir filter . */

static t_class *fir_class;

typedef struct _fir
{
    t_object x_obj; 	/* obligatory header */
    t_float gain;    	/* might be neccesary to do some gain compensation */
	t_float sample_rate;
	float pi, twoPi;    // i won't need these... in case i want to window my impulse
    t_symbol *impulse_array_name;
    t_int   impulse_size;
    t_word *impulse_array;
    float *impulse_response;
    float *input_buffer;
} t_fir;


    /* this is the actual performance routine which acts on the samples.
    It's called with a single pointer "w" which is our location in the
    DSP call list.  We return a new "w" which will point to the next item
    after us.  Meanwhile, w[0] is just a pointer to dsp-perform itself
    (no use to us), w[1] and w[2] are the input and output vector locations,
    and w[3] is the number of points to calculate. */
static t_int *fir_perform(t_int *w)
{
	
	t_fir *x = (t_fir *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);
    
    int i, sample;
    
    for(i = 0; i < x->impulse_size; i++)
    {
        x->impulse_response[i] = x->impulse_array[i].w_float;
    }
    
    for(sample = 0; sample < n; sample++ )
	{
        memcpy(x->input_buffer+1, x->input_buffer,(x->impulse_size - 1) * sizeof(float));
        *(x->input_buffer) = *(in+sample);
        
        *(out+sample) = 0.0;
        for(i = 0; i<x->impulse_size; i++)
            *(out+sample) += *(x->input_buffer+i) * *(x->impulse_response+i);
        *(out+sample) *= x->gain;
	}
	return (w + 5);
}

static void fir_set(t_fir *x, t_symbol *aname)
{
    t_garray *a;
    int i;
    
    if(!(a = (t_garray *)pd_findbyclass(aname, garray_class)))
    {
        pd_error(x, "%s: no such array", aname->s_name);
        x->impulse_size = 0;
    }
    else if(!garray_getfloatwords(a, &x->impulse_size, &x->impulse_array))
    {
        pd_error(x, "%s: bad array data", aname->s_name);
        x->impulse_size = 0;
    }
    else
    {
        x->impulse_array_name = aname;
        garray_usedindsp(a);
        
        // reallocate buffers - if null, same as malloc
        x->impulse_response = (float *)realloc(x->impulse_response, x->impulse_size*sizeof(float));
        x->input_buffer = (float *)realloc(x->input_buffer, x->impulse_size*sizeof(float));
        for(i = 0; i < x->impulse_size; i++)
        {
            x->impulse_response[i] = x->impulse_array[i].w_float;
            x->input_buffer[i] = 0.0;
        }
        post("impulse size %d", x->impulse_size);
    }
}

/* called to start DSP.  Here we call Pd back to add our perform
    routine to a linear callback list which Pd in turn calls to grind
    out the samples. */
static void fir_dsp(t_fir *x, t_signal **sp)
{
	x->sample_rate = sp[0]->s_sr;
    fir_set(x, x->impulse_array_name);
    dsp_add(fir_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

static void *fir_new(t_symbol *selector, int argcount, t_atom *argvec)
{
	long i;
    t_fir *x = (t_fir *)pd_new(fir_class);
	// an inlet to bring in filter coefficients
    outlet_new(&x->x_obj, gensym("signal"));
    x->gain = 1.0f;    	/* might be neccesary to do some gain compensation */
	x->pi = 4.0f * atanf(1.0f);
	x->twoPi = 8.0f * atanf(1.0f);
    x->impulse_response = 0;
    x->input_buffer = 0;
    
    for (i = 0; i < argcount; i++)
    {
        if (argvec[i].a_type == A_FLOAT)
            post("argument to fir~ should be the name of an array");
        else if (argvec[i].a_type == A_SYMBOL)
            fir_set(x, argvec[i].a_w.w_symbol);
    }

	post("this is never going to  work...");
    return (x);
}

// since we allocated some memory, we need a delete function
static void fir_free(t_fir *x)
{
    if(x->impulse_response != 0) free(x->impulse_response);
    if(x->input_buffer != 0) free(x->input_buffer);
}
    /* this routine, which must have exactly this name (with the "~" replaced
    by "_tilde) is called when the code is first loaded, and tells Pd how
    to build the "class". */
void fir_tilde_setup(void)
{
    fir_class = class_new(gensym("fir~"), (t_newmethod)fir_new, (t_method)fir_free,
    	sizeof(t_fir), 0, A_GIMME, 0);
	    /* this is magic to declare that the leftmost, "main" inlet
	    takes signals; other signal inlets are done differently... */
		/* also installs delay_time as the leftmost inlet float */
    CLASS_MAINSIGNALIN(fir_class, t_fir, gain);
    	/* here we tell Pd about the "dsp" method, which is called back
	when DSP is turned on. */
    class_addmethod(fir_class, (t_method)fir_dsp, gensym("dsp"), (t_atomtype)0);
    class_addmethod(fir_class, (t_method)fir_set, gensym("set"), A_GIMME, 0);
}
