#include "m_pd.h"
#include <stdlib.h>
#include <math.h>
#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

/* ------------------------ svf2~ ----------------------------- */

/* a simple one pole filter - getting started . */

static t_class *svf2_class;

typedef struct _svf2
{
    t_object x_obj; 	/* obligatory header */
    t_float frequency;    	/* floats on the first inlet set the frequency */
    t_float oldfrequency;
	t_float	quality;	/* the filter q or quality */
	t_float sample_rate;
	t_float notchOut, lowOut, highOut, bandOut;
} t_svf2;

    /* this is the actual performance routine which acts on the samples.
    It's called with a single pointer "w" which is our location in the
    DSP call list.  We return a new "w" which will point to the next item
    after us.  Meanwhile, w[0] is just a pointer to dsp-perform itself
    (no use to us), w[1] and w[2] are the input and output vector locations,
    and w[3] is the number of points to calculate. */
static t_int *svf2_perform(t_int *w)
{
	t_svf2 *x = (t_svf2 *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);

	// i like counting from zero, so i use sample to count the offset from
	// the start of the in and out blocks
	int sample = 0;
	float c, d, input;

    c = 2.0f * sin(3.141592653589793f * 0.5 * x->frequency/x->sample_rate);
    d = 2.0f * (1.0f - pow(x->quality, 0.25f));
    
    if(c > 0.5f) c = 0.5f;
    if(d > 2.0f) d = 2.0f;
    if(d > (2.0f/c - (c * 0.5f)))
        d = 2.0f/c - (c * 0.5f);

    while(n--)
    {
        input = *(in+sample);
        
        x->notchOut = input - (d * x->bandOut);
        x->lowOut   = x->lowOut + (c * x->bandOut);
        x->highOut  = x->notchOut - x->lowOut;
        x->bandOut  = x->bandOut + (c * x->highOut);
        x->notchOut = input - (d * x->bandOut);
        x->lowOut   = x->lowOut + (c * x->bandOut);
        x->highOut  = x->notchOut - x->lowOut;
        x->bandOut  = x->bandOut + (c * x->highOut);

        *(out+sample) = x->lowOut;
		sample++;
    }
    return (w+5);
}

    /* called to start DSP.  Here we call Pd back to add our perform
    routine to a linear callback list which Pd in turn calls to grind
    out the samples. */
static void svf2_dsp(t_svf2 *x, t_signal **sp)
{
    x->lowOut = x->notchOut = x->highOut = x->bandOut = 0.0f;
	x->sample_rate = sp[0]->s_sr;
    dsp_add(svf2_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

// this input doesn't do anything....
void svf2_quality(t_svf2 *x, t_floatarg g)
{
    x->quality = g;
}

static void *svf2_new(void)
{
    t_svf2 *x = (t_svf2 *)pd_new(svf2_class);
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("quality"));
    outlet_new(&x->x_obj, gensym("signal"));
    x->quality = 0.0f;
    return (x);
}

// since we allocated some memory, we need a delete function
static void svf2_free(t_svf2 *x)
{
}

    /* this routine, which must have exactly this name (with the "~" replaced
    by "_tilde) is called when the code is first loaded, and tells Pd how
    to build the "class". */
void svf2_tilde_setup(void)
{
    svf2_class = class_new(gensym("svf2~"), (t_newmethod)svf2_new, (t_method)svf2_free,
    	sizeof(t_svf2), 0, A_DEFFLOAT, 0);
	    /* this is magic to declare that the leftmost, "main" inlet
	    takes signals; other signal inlets are done differently... */
		/* also installs delay_time as the leftmost inlet float */
    CLASS_MAINSIGNALIN(svf2_class, t_svf2, frequency);
    	/* here we tell Pd about the "dsp" method, which is called back
	when DSP is turned on. */
    class_addmethod(svf2_class, (t_method)svf2_dsp, gensym("dsp"), (t_atomtype)0);
    class_addmethod(svf2_class, (t_method)svf2_quality, gensym("quality"), A_FLOAT, (t_atomtype)0);
}
